# coding:utf-8
"""Views Module."""
import inspect
import time

from tarpit.config import CONFIG as O_O
from tarpit.web import BaseController, route, check_params, Auth
from tarpit.utils.totp import get_totp_token


@route(r'/')
class Account(BaseController):
    """account"""

    @Auth
    async def get(self, *_args, **_kw):

        result = await self.s.user.get_account_info(
            user_id=self.token_args.user_id)

        self.success(result)

    @Auth
    async def put(self, *_args, **_kw):

        @check_params
        def checker(
                name: str = None,
                phone: str = None,
                avatar: str = None,
        ):
            pass

        args = self.parse_json_arguments(checker)

        result = await self.s.user.modify_account_info(
            user_id=self.token_args.user_id,
            name=args.name,
            phone=args.phone,
            avatar=args.avatar,
        )

        self.success(result)


@route(r'/modify_password')
class AccountModifyPassword(BaseController):
    """acccount modify password"""

    @Auth
    async def post(self, *_args, **_kw):

        @check_params
        def checker(user_id: int, oldpass: str, newpass: str):
            pass

        args = self.parse_json_arguments(checker)

        user_info = await self.s.user.check_password_by_id(
            user_id=self.token_args.user_id, password=args.oldpass)

        if not user_info:
            self.fail(401)

        result = await self.s.set_password(
            user_id=args.user_id,
            password=args.password,
        )

        self.success(result)


@route(r'/address')
class AccountAddress(BaseController):
    """account address"""

    @Auth
    async def get(self, *_args, **_kw):

        @check_params
        def checker(address_id: int):
            pass

        args = self.parse_form_arguments(checker)

        result = await self.s.address.get_address_info(
            address_id=args.address_id)

        self.success(result)

    @Auth
    async def post(self, *_args, **_kw):

        @check_params
        def checker(recipient: str, phone: str, address: str):
            pass

        args = self.parse_json_arguments(checker)

        await self.s.address.create_address(
            user_id=self.token_args.user_id,
            recipient=args.recipient,
            phone=args.phone,
            address=args.address)

        self.success()

    @Auth
    async def put(self, *_args, **_kw):

        @check_params
        def checker(
                address_id: int,
                recipient: str = None,
                phone: str = None,
                address: str = None,
        ):
            pass

        args = self.parse_json_arguments(checker)

        await self.s.address.modify_address_info(
            address_id=args.address_id,
            recipient=args.recipient,
            phone=args.phone,
            address=args.address)

        self.success()

    @Auth
    async def delete(self, *_args, **_kw):

        @check_params
        def checker(address_id: int):
            pass

        args = self.parse_form_arguments(checker)

        await self.s.address.delete_address(address_id=args.address_id)

        self.success()


@route(r'/address/list')
class AccountAddressList(BaseController):
    """account list"""

    @Auth
    async def get(self, *_args, **_kw):

        result = await self.s.address.get_address_list(
            user_id=self.token_args.user_id)

        self.success(dict(infos=result))


@route(r'/signin')
class SignIn(BaseController):
    """sign in"""

    async def post(self, *_args, **_kw):

        @check_params
        def checker(username: str, password: str):
            pass

        args = self.parse_form_arguments(checker)

        user_info = await self.s.user.check_password(
            username=args.username, password=args.password)

        if not user_info:
            self.fail(404)
        if user_info.role > 0:
            self.fail(401)

        self.set_token(user_info)
        self.set_header('Access-Control-Expose-Headers',
                        O_O.server.token_header)

        self.success(user_info)



# coding:utf-8
"""Some error class, subclass of `HTTPError`."""

from tornado.web import HTTPError


class ParseJSONError(HTTPError):
    """Exception raised by `BaseHandler.parse_json`.

    This is a subclass of `HTTPError`, so if it is uncaught a 400 response
    code will be used instead of 500 (and a stack trace will not be logged).
    """

    def __init__(self, doc):
        super(ParseJSONError, self).__init__(
            400,
            f'ParseJSONError "{doc}"')



# coding:utf-8
"""Some Verify Function."""
import re

PATTERN = dict(
    email=re.compile(r'^([\w\-.]+)@([\w-]+)(\.([\w-]+))+$'),
    password=re.compile(
        r'^[0-9A-Za-z`~!@#$%^&*()_+\-=\{\}\[\]:;"\'<>,.\\|?/]{6,24}$'))


def check_mail(email):
    """Check Email Pattern."""
    return re.match(PATTERN['email'], email)


def check_password(password):
    """Check Password Pattern."""
    return re.match(PATTERN['password'], password)



# coding:utf-8
"""Module of celery task queue manager."""

import os
import pkgutil
import importlib
import asyncio
import traceback
from functools import wraps

from celery import Celery
from celery.app import trace
from kombu.serialization import register
from sqlalchemy import exc as sqlexc
import pymysql

from tarpit.config import CONFIG as O_O, _ENV as env
from tarpit.dao import transaction
from tarpit.utils import json_encoder, dump_in, dump_out, dump_error

register(
    'tarpit_json',
    json_encoder.thor_dumps,
    json_encoder.thor_loads,
    content_type='application/json',
    content_encoding='utf-8')

trace.LOG_SUCCESS = '\033[0;33mTask [%(id)s] succeeded in %(runtime)1.6ss %(name)s\033[0m'


class CeleryConfig():

    _instance = None
    _app = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(CeleryConfig, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self.queue_list = {
            'user_service': dict(concurrency=3, module='services.user_service')
        }

        self._app = Celery(
            'tasks',
            backend=O_O.celery.backend,
            broker=O_O.celery.broker,
        )

        self._app.conf.update(
            task_serializer='tarpit_json',
            result_serializer='tarpit_json',
            result_expires=180,
            broker_connection_timeout=10,
            broker_pool_limit=10,
            broker_heartbeat=40,
            broker_heartbeat_checkrate=2,
            enable_utc=True,
            timezone='Asia/Hong_Kong',
            task_default_queue='default',
            task_default_exchange='tasks',
            task_default_exchange_type='topic',
            task_default_routing_key='task.default',
            worker_max_memory_per_child=2048,
            worker_log_format=
            '[%(asctime)19.19s %(levelname)1.1s/%(processName)s] %(message)s',
            task_routes={
                f'services.{q}.*': {
                    'queue': f'{env}:{q}'
                } for q in self.queue_list
            },
        )

    @property
    def app(self):
        return self._app


CELERY_CMD = os.getenv('celery_cmd', default='celery')

SERVICES = dict()


class TaskError(Exception):
    """An exception that will turn into an DAO error response.

    :arg int status: Custom DAO status code.
    :arg str message: Message to be written to the log for this error
        (will not be shown to the user unless the `Application` is in debug
        mode).  May contain ``%s``-style placeholders, which will be filled
        in with remaining positional parameters.
    """

    def __init__(self, error):
        try:
            raise error
        except sqlexc.IntegrityError as err:
            self.set_status(1, str(err.orig), err.__class__.__name__)
        except sqlexc.ProgrammingError as err:
            self.set_status(2, str(err.orig), err.__class__.__name__)
        except sqlexc.ResourceClosedError as err:
            self.set_status(3, str(error), err.__class__.__name__)
        except sqlexc.OperationalError as err:
            self.set_status(4, str(err.orig), err.__class__.__name__)
        except UnicodeEncodeError as err:
            self.set_status(5, str(error), err.__class__.__name__)
        except sqlexc.DataError as err:
            self.set_status(6, str(err.orig), err.__class__.__name__)
        except sqlexc.InternalError as err:
            self.set_status(7, str(err.orig), err.__class__.__name__)
        except Exception as err:
            self.set_status(255, str(err), err.__class__.__name__)

    def set_status(self, status, message, error_type):
        self.status = status
        self.message = message
        self.error_type = error_type

    def get_json_object(self):
        return dict(
            status=self.status,
            message=self.message,
            error_type=self.error_type)


def catch_error(function):
    """Wrap a handle shell to a query function."""

    @wraps(function)
    def wrapper(*args, **kwargs):
        """Function that wrapped."""
        try:
            return function(*args, **kwargs)
        except Exception as e:
            return TaskError(e)

    return wrapper


def assemble_celery_cmd(queue):
    cc = CeleryConfig()
    concurrency = cc.queue_list[queue]['concurrency']
    return (f'{CELERY_CMD}', 'worker', '-A', 'worker.app', '-c',
            f'{concurrency}', '-P', 'gevent', '-Q', f'{env}:{queue}', '-n',
            f'{env}:{queue}@%h', f'--loglevel=info')


class Service:

    def __init__(self, mappers=None, celery_params=None):
        self.mappers = mappers or []
        self.celery_params = celery_params or dict()
        self.cur = SERVICES

    def __call__(self, func):

        func = transaction(func, self.mappers)

        func = catch_error(func)

        cc = CeleryConfig()

        if self.celery_params:
            func = cc.app.task(**self.celery_params)(func)
        else:
            func = cc.app.task(func)

        filename = traceback.extract_stack(limit=2)[0].filename
        if '/services' in filename:
            filedir = os.path.dirname(filename).split('/services')[-1]

            modules = filedir.strip('/').split('/')
            filename = os.path.basename(filename)
            filename = os.path.splitext(filename)[0]
            modules.append(filename)

            cur = SERVICES

            for m in modules:
                if not m:
                    continue
                if m not in cur:
                    cur[m] = dict()
                cur = cur[m]
            cur[func.__name__] = func

        return func


def service(mappers=None, celery_params=None, need_transaction=True):
    """Wrapper of 'app.task'."""

    def decorator(func):
        if need_transaction:
            func = transaction(func, mappers or [])

        func = catch_error(func)

        cc = CeleryConfig()
        if celery_params and isinstance(celery_params, dict):
            func = cc.app.task(**celery_params)(func)
        else:
            func = cc.app.task(func)

        filename = traceback.extract_stack(limit=2)[0].filename
        if '/services' in filename:
            filedir = os.path.dirname(filename).split('/services')[-1]

            modules = filedir.strip('/').split('/')
            filename = os.path.basename(filename)
            filename = os.path.splitext(filename)[0]
            modules.append(filename)

            cur = SERVICES

            for m in modules:
                if not m:
                    continue
                if m not in cur:
                    cur[m] = dict()
                cur = cur[m]
            cur[func.__name__] = func

        return func

    return decorator



# coding:utf-8
import sys


def dump(sign, color, show_all, *args):
    sys.stdout.write(f'\n\033[0;{color}m')
    for arg in args:
        if show_all:
            sys.stdout.write('\n' + arg)
        else:
            sys.stdout.write('\n' + arg[:1000])
            if len(arg) > 1000:
                sys.stdout.write(' ...')
    sys.stdout.write(f'\n\033[0m')
    sys.stdout.flush()


def dump_in(*args):
    dump('<', 32, False, *args)


def dump_out(*args):
    dump('>', 33, False, *args)


def dump_error(*args):
    dump('%', 31, True, *args)



# coding:utf-8
"""Views Module."""
import inspect
import time

from tarpit.config import CONFIG as O_O
from tarpit.web import BaseController, route, check_params, Auth, AuthorizationError
from tarpit.task import SERVICES


@route(r'/')
class Index(BaseController):
    """Test index request handler."""

    async def get(self, *_args, **_kwargs):
        """Get method of IndexHandler."""
        self.mg.image.insert_image()
        self.finish(b'<h1>Dejavu!</h1>')


@route(r'*.php')
class PHPScan(BaseController):
    """Test index request handler."""

    async def get(self, *_, **__):
        """return f word."""

        self.finish(b'fuck you')



# coding:utf-8
"""Convert YAML to Python Data."""
import sys
import os
import json

import yaml
from tarpit.entity import Arguments

_ENV = os.getenv('TARPIT_ENV', 'mypc')


class Config:
    """Config Module"""
    _instance = None

    def __new__(cls, **kwargs):
        if not cls._instance:
            cls._instance = super(Config, cls).__new__(cls)
            cls._instance.__init__(**kwargs)
        return cls._instance

    def __init__(self, filepath='settings/config.yml'):
        CONFIG_PATH = os.path.join(os.getcwd(), filepath)
        data = dict(error='Config File Not Found.')
        try:
            with open(CONFIG_PATH, 'r', encoding='utf-8') as config:
                data = yaml.load(config, Loader=yaml.FullLoader)
        except FileNotFoundError:
            pass

        self.data = Arguments(self.convert(data))

    def create(self):
        return self.data

    def convert(self, params):
        new_dict = dict()
        for key in params:
            if isinstance(params[key], dict):
                if _ENV in params[key]:
                    new_dict[key] = params[key][_ENV]
                else:
                    new_dict[key] = self.convert(params[key])
            else:
                new_dict[key] = params[key]
        return new_dict

    def __getattr__(self, name):
        return self.data.__getattr__(name)

    def show(self):
        sys.stdout.write('\nconfig:\n')
        json.dump(self.data.traverse(), sys.stdout, indent=4, sort_keys=True)
        sys.stdout.write('\n\n\n')
        sys.stdout.flush()


CONFIG = Config()


class ErrorCode():

    _code = {}

    def __init__(self, filepath='settings/status.yml'):
        STATUS_PATH = os.path.join(sys.path[0], filepath)
        code_list = []
        try:
            with open(STATUS_PATH, 'r', encoding='utf-8') as config:
                code_list = yaml.load(config, Loader=yaml.FullLoader)
        except FileNotFoundError:
            pass

        if isinstance(code_list, list):
            for item in code_list:
                if 'code' not in item or 'msg' not in item:
                    raise KeyError('Wrong format of status')
                self._code[item.get('code')] = item.get('msg')

    def get_code(self, code, default=None):
        return self._code.get(code, default)


STATUS = ErrorCode()


def get_status_message(code):
    if code == -1:
        return None

    return STATUS.get_code(code, f'Unknown status {code}.')



# coding:utf-8
"""Base module for other views' modules."""
import json
import re
import time
import traceback
import os
import pkgutil
import importlib
import asyncio
from functools import wraps
from collections import abc
from urllib import parse
import inspect

import jwt
from tornado import gen, httpclient
from tornado.web import Finish, RequestHandler, HTTPError
from tornado.log import app_log, gen_log

from tarpit.config import CONFIG as O_O, get_status_message
from tarpit.entity import Arguments, ParseJSONError
from tarpit.utils import dump_in, dump_out, dump_error, json_encoder
from tarpit.task import TaskError, SERVICES
from tarpit.dao import MongoFactory

ROUTES = []


def check_params(func):
    """Decorator to create a parameters checker."""
    parameters = inspect.signature(func).parameters

    for name, value in parameters.items():
        if value.kind == value.VAR_KEYWORD:
            continue
        if value.kind == value.VAR_POSITIONAL:
            continue

        if value.annotation == value.empty:
            raise ValueError(f'Annotation of key <"{name}"> is not exists.')

    def inner(**kwargs):
        for name, value in parameters.items():
            if value.kind == value.VAR_KEYWORD:
                continue
            if value.kind == value.VAR_POSITIONAL:
                continue

            if kwargs.get(name) is None:
                if value.default != value.empty:
                    kwargs[name] = value.default
                else:
                    raise MissingArgumentError(name)

            if kwargs[name] is not None and not isinstance(
                    kwargs[name], value.annotation):
                try:
                    kwargs[name] = value.annotation(kwargs[name])
                except:
                    raise ArgumentTypeError(name, value.annotation)

        return kwargs

    return inner


def route(*path_list: str):
    """Decorator to bind path to a hander."""

    def wrapper(handler: RequestHandler):
        if not issubclass(handler, RequestHandler):
            raise PermissionError('Can\'t routing a nonhandler class.')

        filename = traceback.extract_stack(limit=2)[0].filename
        if '/controllers' in filename:
            filedir = os.path.dirname(filename).split('/controllers')[-1]
            filedir = filedir.strip('/')
            filename = os.path.basename(filename)
            filename = os.path.splitext(filename)[0]
            filename = '' if filename == 'index' else filename
            for path in path_list:
                realpath = path.strip('/')
                realpath = '/' + f'{filedir}/{filename}/{realpath}'.strip('/')
                ROUTES.append((realpath, handler))
        return handler

    return wrapper


def recursive_import(submodule_location, submodule_name=''):
    dirname = submodule_name and (submodule_name + '.')
    module_dict = dict()
    for m in pkgutil.iter_modules(submodule_location):
        if m.ispkg:
            location = f'{submodule_location[0]}/{m.name}'
            recursive_import([location], submodule_name=f'{dirname}{m.name}')
        else:
            module_dict[m.name] = importlib.import_module(f'{dirname}{m.name}')


class Auth:

    def __init__(self, optional=False, **options):
        self.options = options
        self.optional = optional

    def __new__(cls, func=None, **options):
        instance = super(Auth, cls).__new__(cls)
        instance.__init__(**options)
        if func:
            return instance(func)
        else:
            return instance

    def __call__(self, func):
        """Decorator to check authorization of user."""

        def process(ctlr):
            ctlr.set_header('Access-Control-Allow-Origin', '*')
            token_params = Arguments(ctlr.get_token())
            now = int(time.time())
            if not self.optional:
                if not token_params:
                    raise AuthenticationError('Token is empty.')
                if 'timestamp' not in token_params:
                    raise AuthenticationError('No timestamp info in token.')
                if token_params.timestamp < now:
                    raise TokenExpiredError()

                for key in self.options:
                    if isinstance(self.options[key], abc.Iterable):
                        if token_params.get(key) not in self.options[key]:
                            raise AuthorizationError(key)
                    else:
                        if token_params.get(key) != self.options[key]:
                            raise AuthorizationError(key)

            ctlr.token_args = Arguments(token_params)

        @wraps(func)
        async def async_wrapper(ctlr, **kwargs):
            process(ctlr)
            await func(ctlr, **kwargs)

        @wraps(func)
        def wrapper(ctlr, **kwargs):
            process(ctlr)
            func(ctlr, **kwargs)

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return wrapper


class MissingArgumentError(HTTPError):
    """Missing Argument Error."""

    def __init__(self, arg_name):
        super(MissingArgumentError, self).__init__(
            400, f'Missing argument "{arg_name}"')
        self.arg_name = arg_name


class TokenExpiredError(HTTPError):
    """Token Expired Error."""

    def __init__(self):
        super(TokenExpiredError, self).__init__(401, 'Token Expired')


class ArgumentTypeError(HTTPError):
    """Argument Type Error."""

    def __init__(self, arg_name, type_name):
        super(ArgumentTypeError, self).__init__(
            400, f'Argument "{arg_name}" should be {type_name}')


class InternalServerError(HTTPError):
    """Task Error."""

    def __init__(self, task_name, error_msg):
        super(InternalServerError, self).__init__(
            500, f'Error occured when excute task "{task_name}".\n{error_msg}')


class AuthenticationError(HTTPError):
    """Authentication Error."""

    def __init__(self, msg):
        super(AuthenticationError, self).__init__(401, msg)


class AuthorizationError(HTTPError):
    """Authorization Error."""

    def __init__(self, arg_name):
        super(AuthorizationError, self).__init__(
            401, f'User\'s "{arg_name}" not satisfied.')


class Mission:

    def __init__(self, func):
        if not callable(func):
            self.func = self.none_func
            self.func_name = 'none_func'
            raise TypeError(f'Need callable Object, not {type(func)}')
        self.func_name = func.__name__
        self.func = func

    @staticmethod
    def none_func(*args, **kwargs):
        return 'Uncallable function.'

    async def __call__(self, *args, **kwargs):

        if not O_O.worker.mode:
            return self.run(*args, **kwargs)

        async_task = self.func.apply_async(args=args, kwargs=kwargs)

        while not async_task.ready():
            asyncio.sleep(0)

        if async_task.failed():
            dump_error(f'Task Failed: {self.func_name}[{async_task.task_id}]',
                       f'    {str(async_task.result)}')
            raise RuntimeError(str(async_task.result))
        else:
            result = async_task.result

        if isinstance(result, dict) and result.get('error_type'):
            raise RuntimeError(str(async_task.result))
        else:
            return result

    def run(self, *args, **kwargs):

        res = self.func(*args, **kwargs)
        if isinstance(res, TaskError):
            raise InternalServerError(self.func_name, res.get_json_object())

        return res


class ServiceFactory(dict):

    def __init__(self, params=None):
        if isinstance(params, dict):
            super().__init__(params)
        elif not params:
            super().__init__(dict())
        else:
            raise TypeError(
                f"Arguments data should be a 'dict' not {type(params)}.")

    def __getattr__(self, name):
        attr = self.get(name)
        if isinstance(attr, dict):
            return self.__class__(attr)
        else:
            return Mission(attr)

    def __setattr__(self, name, value):
        raise PermissionError('Can not set attribute to <class Arguments>.')


class BaseController(RequestHandler):
    """Custom handler for other views module."""

    def __init__(self, application, request, **kwargs):
        super(BaseController, self).__init__(application, request, **kwargs)
        self.params = Arguments()
        self.token_args = Arguments()
        self.mg = MongoFactory()
        self.s = ServiceFactory(SERVICES)

    def _request_summary(self):
        s = ' '
        return f'{self.request.method.rjust(7, s)} {self.request.remote_ip.rjust(15, s)}  {self.request.path} '

    def log_exception(self, typ, value, tb):
        """Override to customize logging of uncaught exceptions.

        By default logs instances of `HTTPError` as warnings without
        stack traces (on the ``tornado.general`` logger), and all
        other exceptions as errors with stack traces (on the
        ``tornado.application`` logger).

        .. versionadded:: 3.1
        """
        if isinstance(value, HTTPError):
            if value.log_message:
                # format = "%d %s: " + value.log_message
                # args = ([value.status_code,
                #          self._request_summary()] + list(value.args))
                gen_log.warning('\033[0;31m' + value.log_message + '\033[0m')
        else:
            app_log.error(
                "Uncaught exception %s\n%r",
                self._request_summary(),
                self.request,
                exc_info=(typ, value, tb))

    async def options(self, *_args, **_kwargs):
        self.success()

    def prepare(self):
        params = dict(
            device=self.get_argument('device', 'web'),
            lang=self.get_argument('lang', 'cn').lower(),
            remote_ip=self.request.remote_ip,
            request_time=time.time())
        self.params = Arguments(params)

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Methods', 'GET,POST,PUT,DELETE')
        self.set_header('Access-Control-Allow-Headers',
                        'Authorization,Content-Type,Thor-Token')

    def get_token(self):
        header_name = O_O.server.token_header or 'Thor-Token'
        token = self.request.headers.get(header_name)
        try:
            return jwt.decode(token, O_O.server.token_secret)
        except jwt.DecodeError:
            return None

    def set_token(self, token_params: dict = None, expired: int = 3600):
        token_params = token_params or dict()
        token = jwt.encode(
            dict(token_params, timestamp=int(time.time()) + expired),
            O_O.server.token_secret)
        self.set_header(O_O.server.token_header or 'Thor-Token', token)

    def parse_form_arguments(self, checker=None):
        """Parse FORM argument like `get_argument`."""
        if O_O.debug:
            dump_in(f'Input: {self.request.method} {self.request.path}',
                    self.request.body.decode()[:500])

        args = {
            k: v[0].decode() for k, v in self.request.arguments.items() if v[0]
        }

        if checker:
            args = checker(**args)

        return Arguments(args)

    def parse_json_arguments(self, checker=None):
        """Parse JSON argument like `get_argument`."""
        if O_O.debug:
            dump_in(f'Input: {self.request.method} {self.request.path}',
                    self.request.body.decode()[:500])

        try:
            args = json.loads(self.request.body.decode('utf-8'))
        except json.JSONDecodeError as exception:
            dump_error(self.request.body.decode())
            raise ParseJSONError(exception.doc)

        if not isinstance(args, dict):
            dump_error(self.request.body.decode())
            raise ParseJSONError('Request body should be a dictonary.')

        if checker:
            args = checker(**args)

        return Arguments(args)

    def finish_with_json(self, data):
        """Turn data to JSON format before finish."""
        self.set_header('Content-Type', 'application/json')

        if O_O.debug:
            if self.request.method == 'POST':
                info_list = [
                    f'\033[0;33mOutput: {self.request.method} {self.request.path}'
                ]
                if self.request.query:
                    query_list = [
                        f'\033[0;33m{i[0]:15s} {i[1]}'
                        for i in parse.parse_qsl(self.request.query)
                    ]
                    info_list.append('\n' + '\n'.join(query_list))
                if self.request.body:
                    try:
                        info_list.append('\n\033[0;33m' +
                                         self.request.body.decode())
                    except UnicodeDecodeError:
                        pass
                if data:
                    info_list.append('\n\033[0;33m' + json.dumps(data))
                dump_out(*info_list)

        raise Finish(json.dumps(data).encode())

    def fail(self, status, data=None, polyfill=None, **_kwargs):
        """assemble and return error data."""
        msg = get_status_message(status)
        self.finish_with_json(
            dict(status=status, msg=msg, data=data, **_kwargs))

    def success(self, data=None, msg='Successfully.', **_kwargs):
        """assemble and return error data."""
        self.finish_with_json(dict(status=0, msg=msg, data=data))

    async def fetch(self, api, method='GET', body=None, headers=None,
                    **_kwargs):
        """Fetch Info from backend."""
        body = body or dict()

        _headers = dict(host=self.request.host)
        if headers:
            _headers.update(headers)

        if '://' not in api:
            api = f'http://{O_O.server.back_ip}{api}'

        back_info = await httpclient.AsyncHTTPClient().fetch(
            api,
            method=method,
            headers=_headers,
            body=json.dumps(body),
            raise_error=False,
            allow_nonstandard_methods=True)

        res_body = back_info.body and back_info.body.decode() or None

        if back_info.code >= 400:
            return Arguments(
                dict(http_code=back_info.code, res_body=res_body, api=api))

        try:
            return Arguments(json.loads(res_body))
        except json.JSONDecodeError:
            raise ParseJSONError(res_body)



# coding:utf-8
from functools import wraps
import traceback
import enum
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Query, Session
from motor import MotorGridFSBucket, motor_tornado as motor
from pymongo import MongoClient

from tarpit.utils.logger import dump_in, dump_out, dump_error
from tarpit.config import CONFIG as O_O


class MongoBase():
    """Mongo Client Set."""

    __collection_name__ = None
    __alias__ = None

    def __init__(self):
        T_T = O_O.database.mongo
        self.client = motor.MotorClient(T_T.client).__getattr__(T_T.db)
        self.sync_client = MongoClient(T_T.client).__getattr__(T_T.db)
        self.__setattr__(self.__collection_name__,
                         self.client.__getattr__(self.__collection_name__))

    def __collection_init__(self):
        """Write some collection initial operation, like set some index."""
        pass


class MongoFactory():
    """Mongo Client Set."""

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(MongoFactory, cls).__new__(cls)
            cls._instance.mongo_init()
            print('mongo new')

        return cls._instance

    @classmethod
    def mongo_init(self):
        self.collection_dict = {(scls.__alias__ or scls.__collection_name__):
                                scls for scls in MongoBase.__subclasses__()}
        for scls in MongoBase.__subclasses__():
            scls().__collection_init__()

    def __getattr__(self, name):
        return self.collection_dict[name]()


class SessionMaker:

    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(SessionMaker, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self.new_engine = create_engine(
            O_O.database.mysql,
            echo=False,
            pool_recycle=1000,
            encoding='utf-8',
            isolation_level='REPEATABLE_READ')
        self.session_maker = sessionmaker(bind=self.new_engine)

    def create(self):
        return self.session_maker()


class BaseMapper:
    """Base Mapper."""

    __alias__ = None

    def __init__(self, session):
        if not isinstance(session, Session):
            raise ValueError(
                "session is not instance of sqlalchemy.orm.session.Session.")
        self.session = session


def dict_of(entity, *options, **alias):
    if entity is None:
        return dict()

    res = dict()
    for key in options:
        alias[key] = None

    for key in entity.__dict__:
        if not alias or key in alias:
            if not key.startswith('_'):
                if isinstance(alias.get(key), str):
                    real_key = alias[key]
                else:
                    real_key = key
                if isinstance(entity.__dict__[key], enum.Enum):
                    res[real_key] = entity.__dict__[key].name
                elif isinstance(entity.__dict__[key], datetime):
                    res[real_key] = int(entity.__dict__[key].timestamp())
                else:
                    res[real_key] = entity.__dict__[key]

    return res


class DaoContext:

    def __init__(self, session, mapper):
        self.session = session
        self.mapper = mapper

    def get_mapper(self):
        return self.mapper(self.session)

    def commit(self):
        self.session.commit()


def transaction(function, mappers, rollback=True):
    """Wrap a handle shell to a query function."""

    @wraps(function)
    def wrapper(*args, **kwargs):
        """Function that wrapped."""
        session = SessionMaker().create()
        if mappers:
            mapper_dict = {(scls.__alias__ or scls.__name__): scls
                           for scls in BaseMapper.__subclasses__()}
            parents = tuple(mapper_dict[m] for m in mappers if m in mapper_dict)
            mapper = type('Mapper', parents, {})
        else:
            mapper = BaseMapper

        context = DaoContext(session, mapper)

        try:
            return function(context, *args, **kwargs)
        except Exception as e:
            session.rollback()
            raise e
        finally:
            session.close()

    return wrapper


# coding:utf-8
"""Lazor Database Module."""
import uuid
import enum
from datetime import datetime
from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import (CHAR, BigInteger, Column, Enum, Integer, SmallInteger,
                        Text, String, Numeric, Float, TIMESTAMP)
from sqlalchemy.dialects.mysql import MEDIUMTEXT, INTEGER, DECIMAL
from sqlalchemy import PrimaryKeyConstraint, Sequence, UniqueConstraint, Index
from sqlalchemy import text

from .sqlalchemy_base import Base, CreatedAt, UpdatedAt, Remark, Deleted


class User(Base, CreatedAt, UpdatedAt, Remark, Deleted):
    """用户"""
    __tablename__ = 'user'

    user_id = Column(Integer)
    name = Column(String(30))
    phone = Column(String(100))
    password = Column(String(36))
    avatar = Column(String(255))

    __table_args__ = (
        PrimaryKeyConstraint('user_id'),
        Index('ix_user_phone', phone),
        Index('ix_user_name', name),
        Index('ix_user_password', password),
        {
            'mysql_engine': 'InnoDB'
        },
    )



# coding:utf-8
"""Some function to generate code."""
import random
import time
import string
import hmac

from tarpit.config import CONFIG as O_O

pool = string.ascii_letters + string.digits


def rand6():
    return ''.join(random.choice(pool) for _ in range(6))


def rand12():
    rand = ''.join(random.choice(pool) for _ in range(8))
    stamp = hex(int(time.time()))[-4:]
    return rand + stamp


def generate_id():
    stamp = hex(int(time.time()))[2:]
    rand = ''.join(random.choice(pool) for _ in range(8))
    return rand + stamp


def encode_passwd(passwd):
    return hmac.new(O_O.server.pass_mixin.encode('utf-8'),
                    passwd.encode()).hexdigest()


def encode_msg(code, timestamp, user_id):
    return hmac.new(code.encode(),
                    (str(timestamp) + str(user_id)).encode()).hexdigest()



import base64
import struct
import hmac
import hashlib
import time
import random


def get_hotp_token(secret, intervals_no):
    key = base64.b32decode(secret, True)
    msg = struct.pack(">Q", intervals_no)
    h = hmac.new(key, msg, hashlib.sha1).digest()
    o = h[19] & 15
    h = (struct.unpack(">I", h[o:o + 4])[0] & 0x7fffffff) % 1000000
    return f'{h:06d}'


def get_totp_token(secret):
    return get_hotp_token(secret, intervals_no=int(time.time()) // 30)

def get_totp_secret():
    return ''.join(random.choices('ABCDEFGHIJKLMNOPQRSTUVWXYZ234567', k=16))



#!/usr/bin/env python
# coding: utf-8

from setuptools import setup, find_packages

setup(
    name='tarpit',
    version='0.0.4',
    author='plank',
    author_email='sieglive@gmail.com',
    url='https://lazor.cn',
    description=u'吃枣药丸',
    license="GPLv3+",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'alembic==1.0.7',
        'celery==4.2.1',
        'gevent==1.4.0',
        'motor==2.0.0',
        'PyJWT==1.7.1',
        'pymongo==3.7.2',
        'PyMySQL==0.9.3',
        'PyYAML==5.1',
        'redis==3.2.0',
        'requests==2.21.0',
        'SQLAlchemy==1.2.18',
        'tornado==5.1.1',
        'cryptography==2.6.1',
    ],
    entry_points={'console_scripts': ['tarpit=tarpit:main']},
)



from sqlalchemy.orm import Query
from sqlalchemy import or_

from database.sqlalchemy_models import User
from tarpit.dao import BaseMapper


class UserMapper(BaseMapper):
    """User Query Factory."""

    __alias__ = 'user'

    def insert_user(self, name, phone, password, role, avatar, **kwargs):
        user = User(
            name=name, phone=phone, password=password, role=role, avatar=None)
        self.session.add(user)
        return user

    def update_user_name(self, user_id, name, **kwargs):
        query = self.session.query(User).filter(User.user_id == user_id)
        return query.update(dict(name=name))

    def update_user_phone(self, user_id, phone, **kwargs):
        query = self.session.query(User).filter(User.user_id == user_id)
        return query.update(dict(phone=phone))

    def update_user_avatar(self, user_id, avatar, **kwargs):
        query = self.session.query(User).filter(User.user_id == user_id)
        return query.update(dict(avatar=avatar))

    def update_user_password(self, user_id, password, **kwargs):
        query = self.session.query(User).filter(User.user_id == user_id)
        return query.update(dict(password=password))

    def query_user_by_id(self, user_id, **kwargs):
        query = self.session.query(User)
        query = query.filter(User.user_id == user_id)
        return query.first()


# coding:utf-8
"""Library Module."""
import os
import shutil
import sys
import subprocess
import argparse

from tornado import ioloop, gen
from tornado.web import Application
from tornado.options import options
from tornado.log import enable_pretty_logging
import yaml

from tarpit.config import CONFIG as O_O
from tarpit.web import ROUTES, recursive_import
# from tarpit.task import CeleryConfig, assemble_celery_cmd
from tarpit.utils import json_encoder


def set_dependence(module):
    recursive_import(module.__spec__.submodule_search_locations,
                     module.__name__)


def generate(dst):
    src = os.path.join(os.path.dirname(__file__), 'templates')
    shutil.copytree(src, dst)


def serve():
    enable_pretty_logging()

    for r in ROUTES:
        print(f'{r[0]:30s} {r[1]}')

    app = Application(handlers=ROUTES, **O_O.application)
    app.listen(O_O.server.port, **O_O.httpserver)

    O_O.show()
    print('\nstart listen...')
    ioloop.IOLoop.instance().start()


def main():
    sys.path.append(os.getcwd())

    parser = argparse.ArgumentParser(description='Tarpit commend line tools.')
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')

    subparsers.add_parser('serve')

    worker_parser = subparsers.add_parser('worker')
    # cc = CeleryConfig()
    # worker_parser.add_argument(
    #     '-q',
    #     '--queue',
    #     type=str,
    #     choices=[*cc.queue_list],
    #     help='Run worker instance.')

    create_parser = subparsers.add_parser('create')
    create_parser.add_argument('directory', type=str, help='target directory.')

    params = parser.parse_args()

    if params.command != 'create':
        import mappers
        import services
        import controllers
        import collection
        set_dependence(collection)
        set_dependence(mappers)
        set_dependence(services)
        set_dependence(controllers)

    if params.command == 'serve':
        serve()
    elif params.command == 'worker':
        print('runworker')
    elif params.command == 'create':
        if os.path.exists(params.directory):
            parser.error('directory exists')
        generate(params.directory)
    else:
        parser.error('No valid options.')


# coding:utf-8
"""Arguments class, a wrapper of dict."""
import sys
import json


class Arguments(dict):
    """Class to manage arguments of a requests."""

    def __init__(self, params=None):
        if isinstance(params, dict):
            super().__init__(params)
        elif not params:
            super().__init__(dict())
        else:
            raise TypeError(
                f"Arguments data should be a 'dict' not {type(params)}.")

    def __getattr__(self, name):
        attr = self.get(name)
        if isinstance(attr, dict):
            attr = self.__class__(attr)
        return attr

    def __setattr__(self, name, value):
        if self.__getattr__(name):
            raise PermissionError('Can not set attribute to <class Arguments>.')
        else:
            super().__setattr__(name, value)

    def insert(self, key, value):
        """Add a variable to args."""
        if key in self:
            raise PermissionError(f'Key {key} is already exists.')
        else:
            self[key] = value

    def traverse(self, pre=''):
        new_dict = dict()
        for key in self:
            if isinstance(self[key], dict):
                new_dict.update(self.__getattr__(key).traverse(pre + key + '.'))
            else:
                new_dict[pre + key] = self[key]
        return new_dict